package altro.lockEvoluto_Sep17;

import java.util.Date;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/*
 * Si noti che i metodi di StepCondition non sono protetti da lock perchè si assume che una condition sia utilizzata
 * mentre si ha in possesso il lock di appartenenza (in questo caso è theLock)
 */
public class StepCondition implements Condition
{
	/**
	 * Campo utilizzato per memorizzare il Lock a cui questa condition appartiene
	 */
	private Lock theLock;
	/**
	 * Si usa una condition normale incapsulata per realizzare la stepcondition
	 */
	private Condition inCondition;
	int contaSignal;
	StepCondition(Lock l)
	{
		theLock = l;
		inCondition = theLock.newCondition();
		contaSignal = 0;
	}
	void await(int volte) throws InterruptedException
	{
		while (volte > 0)
		{
			inCondition.await();
			/**
			 * Smaltisco eventuali signal che ho conteggiato
			 */
			if (contaSignal > 0)
			{   
				int temp = volte;
				volte -= contaSignal;
				if (contaSignal >= temp) 
					contaSignal -= temp;
				else
					contaSignal = 0;
			}
		}
	}
	void signal(int volte)
	{
		while (--volte > 0) {
			contaSignal++;
			inCondition.signal();
		}
	}
	@Override
	public void await() throws InterruptedException {
		inCondition.await();
		/**
		 * Si assume che se contaSignal > 0, allora ci sono delle signal non ancora smaltite, e
		 * dunque decremento contaSignal per conteggiare di aver smaltito 1 (UNA) signal.
		 * Se contaSignal == 0 e ci troviamo ad eseguire questo punto del codice allora siamo nel caso
		 * di una spurious wake-up.
		 * Vale anche per awaitUninterruptibly, awaitNanos, ecc.
		 */
		if ( contaSignal > 0 ) contaSignal--;
	}
	@Override
	public void awaitUninterruptibly() {
		inCondition.awaitUninterruptibly();
		if (contaSignal > 0) contaSignal--;
	}
	@Override
	public long awaitNanos(long nanosTimeout) throws InterruptedException {
		long retValue = inCondition.awaitNanos(nanosTimeout);
		if(contaSignal > 0) contaSignal--;
		return retValue;
	}
	@Override
	public boolean await(long time, TimeUnit unit) throws InterruptedException {
		boolean retValue = inCondition.await(time, unit);
		if(contaSignal > 0) contaSignal--;
		return retValue;
	}
	@Override
	public boolean awaitUntil(Date deadline) throws InterruptedException {
		boolean retValue = inCondition.awaitUntil(deadline);
		if(contaSignal > 0) contaSignal--;
		return retValue;
	}
	@Override
	public void signal() {
		contaSignal++;
		inCondition.signal();
	}
	@Override
	public void signalAll() {
		contaSignal++;
		inCondition.signalAll();
	}
}
